lrm.las <- function(X, y, group, lambda, beta, mu,
                    eps.outer, eps.inner, max.it, optim.func){
  
  ## INITIALIZATION
  ## Managing intercept
  intercept <- rep(FALSE,ncol(X))
  if (length(mu)>0) {
    intercept <- c(TRUE,intercept)
    beta <- c(mu,beta)
    X <- cbind(1,X)
    lambda <- c(0,lambda)
  }
  active <- abs(beta) > 0 | intercept
  p  <- ncol(X)
  L0 <- 1
  iter.optim <- c()
  ## Check if optimality is reached at starting point
  dL.min  <- dL.lasso.binomial(beta,X,y,lambda)
  gap.old <- max(abs(dL.min))
  if (gap.old < eps.outer*p) {
    return(list(beta = beta[!intercept],
                mu   = beta[intercept],
                iter.optim=iter.optim,gap=gap.old,status="converged"))
  } else {
    ## UPDATE THE ACTIVE SET / ACTIVE GROUP
    new <- intersect(which.max(abs(dL.min)),which(!intercept))
    if (all(active[new])) {
      eps.optim  <- eps.outer
    } else {
      active[new] <- TRUE
      eps.optim  <- eps.inner
    }
  }

  while (1) {
    
    ## (1) OPTIMIZATION OVER A
    ## _____________________________________________________________

    res <- optim.func(beta[active], X[, active],
                      y, lambda[active], eps.optim, L0)
    L0 <- res$L
    iter.optim <- c(iter.optim,res$i)
    beta[active | intercept] <- res$xk
    dL.min[active] <- dL.lasso.binomial(beta[active], X[,active],y,
                                        lambda[active])
    
    ## (2) TEST LAGRAGIAN VIOLATION AND GROUP DELETION IF APPLICABLE
    ## _____________________________________________________________
    go.zero <- abs(beta) < eps.optim & active & !intercept
    if (sum(go.zero) > 0) {
      zeroed <- go.zero & abs(dL.min) < eps.optim
      dL.min[zeroed] <- 0
      active[zeroed] <- FALSE
      beta[zeroed]   <- 0
    }
    
    ## (3) OPTIMALITY TESTING AND GROUP ACTIVATION IF APPLICABLE
    ## _____________________________________________________________
    gap <- max(abs(dL.min))    
    if (length(iter.optim) == max.it) {
      return(list(beta = beta[!intercept],
                  mu   = beta[intercept],
                  iter.optim=iter.optim,
                  status="max # of iterates reached",gap=gap))
    }
    if (gap < eps.outer*p) {
      return(list(beta = beta[!intercept],
                  mu   = beta[intercept],
                  iter.optim=iter.optim,gap=gap,status="converged"))
    } else {
      ## UPDATE THE ACTIVE SET / ACTIVE GROUP
      new <- intersect(which.max(abs(dL.min)),which(!intercept))
      if (all(active[new])) {
        eps.optim <- eps.outer
      } else {
        active[new] <- TRUE
        eps.optim <- eps.inner
      }
    }
  }
}

L.lasso.binomial <- function(beta, X, y, lambda) {

  eta <- crossprod(t(X), beta)
  L  <- - sum( y * eta - log(1 + exp(eta)) ) + sum(lambda * abs(beta))

  return(L)
}

dL.lasso.binomial <- function(beta,X,y,lambda) {

  zero <- abs(beta) == 0
  dL <- rep(0,length(beta))

  eta <- crossprod(t(X),beta)
  nabla.f <- - crossprod(X,y-1/(1+exp(-eta)))
  n1.nabla.f <- abs(nabla.f)
  
  ## terms whose norm1 is different from zero
  dL[!zero] <- nabla.f[!zero] + lambda[!zero] * sign(beta)[!zero]
  
  ## terms whose norm2 is zero
  to.zero    <- n1.nabla.f[zero] <= lambda[zero]
  
  ## which ones can be null ?
  dL[zero][to.zero] <- 0
  dL[zero][!to.zero] <- nabla.f[zero][!to.zero] * (1 - lambda[zero][!to.zero]/n1.nabla.f[zero][!to.zero])

  return(dL)
}  

lrm.grp <- function(X, y, group, lambda, beta, mu,
                    eps.outer, eps.inner, max.it, optim.func){
  
  ## INITIALIZATION
  intercept <- rep(FALSE,ncol(X))
  ## Managing intercept
  if (length(mu)>0) {
    intercept <- c(TRUE,intercept)
    beta <- c(mu,beta)
    X <- cbind(1,X)
    lambda <- c(0,lambda)
    group  <- c(0,group) + 1
  }
  pk <- tabulate(factor(group))
  active <- group.norm.rep(beta, pk) > 0 | intercept
  p  <- ncol(X)
  L0 <- 1
  iter.optim <- c()
  ## Check if optimality is reached at starting point
  dL.min <- dL.group.binomial(beta,X,y,lambda,pk)
  n2.dL  <- group.norm(dL.min,pk)
  gap.old <- max(n2.dL)
  if (gap.old < eps.outer*p) {
    return(list(beta = beta[!intercept],
                mu   = beta[intercept],
                iter.optim=iter.optim,gap=gap.old,status="converged"))
  } else {
    ## UPDATE THE ACTIVE SET / ACTIVE GROUP
    new <- setdiff(which.max(n2.dL), sum(intercept))
    if (all(active[group == new])) {
      eps.optim  <- eps.outer
    } else {
      active[group == new] <- TRUE
      eps.optim  <- eps.inner
    }
  }
  
  while (1) {
    
    ## (1) OPTIMIZATION OVER A
    ## _____________________________________________________________
    res <- optim.func(beta[active], X[,active], y, lambda[active],
                      tabulate(factor(group[active])), eps.optim, L0)
    L0 <- res$L
    iter.optim <- c(iter.optim,res$i)
    beta[active | intercept] <- res$xk
    dL.min[active] <- dL.group.binomial(beta[active],X[,active],y,
                                        lambda[active],
                                        tabulate(factor(group[active])))
    
    ## (2) TEST LAGRAGIAN VIOLATION AND GROUP DELETION IF APPLICABLE
    ## _____________________________________________________________
    go.zero <- intersect(which(group.norm(beta,pk) < eps.optim),group[active])
    go.zero <- setdiff(go.zero, sum(intercept))
    if (length(go.zero) > 0) {
      n2.dL  <- group.norm.rep(dL.min,pk)
      zeroed <- group %in% go.zero & n2.dL < eps.optim
      dL.min[zeroed] <- 0
      active[zeroed] <- FALSE
      beta[zeroed]   <- 0
    }

    ## (3) OPTIMALITY TESTING AND GROUP ACTIVATION IF APPLICABLE
    ## _____________________________________________________________
    n2.dL   <- group.norm(dL.min,pk)
    gap <- max(n2.dL)
    if (length(iter.optim) == max.it) {
      return(list(beta = beta[!intercept],
                  mu   = beta[intercept],
                  iter.optim=iter.optim,
                  status="max # of iterates reached",gap=gap))
    }    
    if (gap < eps.outer*p) {
      return(list(beta = beta[!intercept],
                  mu   = beta[intercept],
                  iter.optim=iter.optim,gap=gap,status="converged"))
    } else {
      ## UPDATE THE ACTIVE SET / ACTIVE GROUP
      new <- setdiff(which.max(n2.dL), sum(intercept))
      if (all(active[group == new])) {
        eps.optim <- eps.outer
      } else {
        active[group == new] <- TRUE
        eps.optim <- eps.inner
      }
    }
  }
}

L.group.binomial <- function(beta,X,y,lambda,pk) {

  eta <- crossprod(t(X), beta)
  n2  <- group.norm.rep(beta,pk) / rep.int(pk,pk)
  L   <- - sum( y * eta - log(1 + exp(eta)) ) + sum(lambda *n2)

  return(L)
}

dL.group.binomial <- function(beta,X,y,lambda,pk) {

  eta <- crossprod(t(X),beta)
  nabla.f <- - crossprod(X, y-1/(1+exp(-eta)) )
  n2.nabla.f <- group.norm.rep(nabla.f,pk)
  n2 <- group.norm.rep(beta,pk)
  
  zero <- n2 == 0
  ## terms whose norm2 is different from zero
  dL <- rep(0,length(beta))
  dL[!zero] <- nabla.f[!zero] + lambda[!zero] * beta[!zero] / n2[!zero]
  
  ## terms whose norm2 is zero
  to.zero    <- n2.nabla.f[zero] <= lambda[zero]
  
  ## which ones can be null ?
  dL[zero][to.zero] <- 0
  dL[zero][!to.zero] <- nabla.f[zero][!to.zero] * (1 - lambda[zero][!to.zero]/n2.nabla.f[zero][!to.zero])

  return(dL)
}  

lrm.coo <- function(X, y, group, lambda, beta, mu,
                    eps.outer, eps.inner, max.it, optim.func){

  ## INITIALIZATION
  intercept <- rep(FALSE,ncol(X))
  ## Managing intercept
  if (length(mu)>0) {
    intercept <- c(TRUE,intercept)
    beta <- c(mu,beta)
    X <- cbind(1,X)
    lambda <- c(0,lambda)
    group  <- c(0,group) + 1
  }
  L0 <- 1
  p  <- ncol(X)
  pk <- tabulate(factor(group))
  ## active <- group.norm.rep(beta, pk) > 0 | intercept
  active.pos <- group.norm.rep(pmax.int(0, beta),pk) > 0 | intercept
  active.neg <- group.norm.rep(pmax.int(0,-beta),pk) > 0 | intercept
  iter.optim <- c()
  ## Check if optimality is reached at the starting point
  dL.min <- dL.coop.binomial(beta,X,y,lambda,pk)
  n2.dL.pos  <- group.norm(pmax.int(0, dL.min),pk)
  n2.dL.neg  <- group.norm(pmax.int(0,-dL.min),pk)
  pmax.n2.dL <- pmax(n2.dL.pos,n2.dL.neg)
  gap.old <- max(pmax.n2.dL)
  if (gap.old < eps.outer*p) {
    return(list(beta = beta[!intercept],
                mu   = beta[intercept],
                iter.optim=iter.optim,gap=gap.old,status="converged"))
  } else {
    ## UPDATE THE ACTIVE SET / ACTIVE GROUP
    new <- setdiff(which.max(pmax.n2.dL), sum(intercept))
    if (all(active.neg[group == new]) | all(active.pos[group == new])) {
      eps.optim <- eps.outer
    } else {
      eps.optim <- eps.inner
    }
    if (length(new) > 0) { ## if there is just the intercept to optimize
      if (n2.dL.pos[new] > n2.dL.neg[new]) {
        active.neg[group == new] <- TRUE
      } else {
        active.pos[group == new] <- TRUE
      }
    }
  }
  

  while (1) {

    ## (1) OPTIMIZATION OVER A
    ## _____________________________________________________________
    lower.bound <- rep(-Inf,p)
    upper.bound <- rep(+Inf,p)
    lower.bound[active.pos & !active.neg & !intercept] <- 0
    upper.bound[active.neg & !active.pos & !intercept] <- 0
    active <- active.pos | active.neg

    res <- optim.func(beta[active], X[,active], y, lambda[active],
                      tabulate(factor(group[active])), eps.optim,
                      lower = lower.bound[active], upper = upper.bound[active], L0)
    L0 <- res$L
    iter.optim <- c(iter.optim,res$i)
    beta[active] <- res$xk

    dL.min[active] <- dL.coop.binomial(beta[active],X[,active],y,
                                       lambda[active],
                                       tabulate(factor(group[active])))
    
    ## (2) TEST LAGRAGIAN VIOLATION AND GROUP DELETION IF APPLICABLE
    ## _____________________________________________________________
    n2.beta.pos <- group.norm(pmax.int(0,beta),pk)
    go.zero.pos <- intersect(which(n2.beta.pos < eps.optim),group[active.pos])
    go.zero.pos <- setdiff(go.zero.pos, sum(intercept))
    if (length(go.zero.pos) > 0) {
      n2.dL.neg  <- group.norm.rep(pmax.int(0,-dL.min),pk)
      zeroed <- group %in% go.zero.pos & n2.dL.neg < eps.optim & beta > 0
      dL.min[zeroed] <- 0
      active.pos[zeroed] <- FALSE
      beta[zeroed]       <- 0
    }

    n2.beta.neg <- group.norm(pmax.int(0,-beta),pk)
    go.zero.neg <- intersect(which(n2.beta.neg < eps.optim),group[active.neg])
    go.zero.neg <- setdiff(go.zero.neg, sum(intercept))
    if (length(go.zero.neg) > 0) {
      n2.dL.pos  <- group.norm.rep(pmax.int(0,dL.min),pk)
      zeroed <- group %in% go.zero.neg & n2.dL.pos < eps.optim & beta < 0
      dL.min[zeroed] <- 0
      active.neg[zeroed] <- FALSE
      beta[zeroed]       <- 0
    }
    
    ## (3) OPTIMALITY TESTING AND GROUP ACTIVATION IF APPLICABLE
    ## _____________________________________________________________
    ## Get the entry for which the constraint is the most violated
    n2.dL.pos  <- group.norm(pmax.int(0, dL.min),pk)
    n2.dL.neg  <- group.norm(pmax.int(0,-dL.min),pk)
    pmax.n2.dL <- pmax.int(n2.dL.pos,n2.dL.neg)
    gap        <- max(pmax.n2.dL)
    if (length(iter.optim) == max.it) {
      return(list(beta = beta[!intercept],
                  mu   = beta[intercept],
                  iter.optim=iter.optim,status="max # of iterates reached",
                  gap=gap))
    }
##     if (abs(gap - gap.old) < eps.outer) {
##       return(list(beta = beta[!intercept],
##                   mu   = beta[intercept],
##                   iter.optim=iter.optim,status="no further improvement...",
##                   gap=gap))
##     }
    
    if (gap < eps.outer*p) {
      return(list(beta = beta[!intercept],
                  mu   = beta[intercept],
                  iter.optim=iter.optim,gap=gap,status="converged"))
    } else {
      gap.old <- gap
      ## UPDATE THE ACTIVE SET / ACTIVE GROUP
      new <- setdiff(which.max(pmax.n2.dL), sum(intercept))
      if (all(active.neg[group == new]) | all(active.pos[group == new])) {
        eps.optim <- eps.outer
      } else {
        eps.optim <- eps.inner
      }
      if (length(new) > 0) {
        if (n2.dL.pos[new] > n2.dL.neg[new]) {
          active.neg[group == new] <- TRUE
        } else {
          active.pos[group == new] <- TRUE
        }
      }
    }
  }
}

L.coop.binomial <- function(beta,X,y,lambda,pk) {

  eta <- crossprod(t(X), beta)
  a.pos <- group.norm.rep(pmax.int(0, beta),pk)/rep.int(pk,pk)
  a.neg <- group.norm.rep(pmax.int(0,-beta),pk)/rep.int(pk,pk)
  
  L <- - sum( y * eta - log(1 + exp(eta)) ) + sum(lambda *(a.pos + a.neg))

  return(L)
}

dL.coop.binomial <- function(beta,X,y,lambda,pk) {

  eta <- crossprod(t(X), beta)
  nabla.f <- - crossprod(X, y-1/(1+exp(-eta)) )
  a.pos <- group.norm.rep(pmax.int(0, beta),pk)
  a.neg <- group.norm.rep(pmax.int(0,-beta),pk)
  n2.theta.pos <- group.norm.rep(pmax.int(0, nabla.f),pk)
  n2.theta.neg <- group.norm.rep(pmax.int(0,-nabla.f),pk)
      
  ## manage cases where alpha+ > 0 or alpha- < 0
  dL <- nabla.f
  neg <- beta < 0
  pos <- beta > 0
  dL[pos] <- nabla.f[pos] + lambda[pos]*pmax.int(rep(0,sum(pos)), beta[pos]) / a.pos[pos]
  dL[neg] <- nabla.f[neg] - lambda[neg]*pmax.int(rep(0,sum(neg)),-beta[neg]) / a.neg[neg]
  
  ## manage cases where beta = 0 and (alpha- = 0 and/or alpha+ = 0)
  zero.pos  <- abs(beta) == 0 & nabla.f > 0 & a.neg == 0
  zero.neg  <- abs(beta) == 0 & nabla.f < 0 & a.pos == 0
  
  pos.to.zero  <- n2.theta.pos[zero.pos] <= lambda[zero.pos]
  neg.to.zero  <- n2.theta.neg[zero.neg] <= lambda[zero.neg]

  dL[zero.pos][!pos.to.zero] <- nabla.f[zero.pos][!pos.to.zero] * (1-lambda[zero.pos][!pos.to.zero]/n2.theta.pos[zero.pos][!pos.to.zero])
  dL[zero.neg][!neg.to.zero] <- nabla.f[zero.neg][!neg.to.zero] * (1-lambda[zero.neg][!neg.to.zero]/n2.theta.neg[zero.neg][!neg.to.zero])
  
  dL[zero.pos][pos.to.zero] <- 0
  dL[zero.neg][neg.to.zero] <- 0
  
  return(dL)
}
